<?php

namespace Drupal\sdc\Component;

use Drupal\Core\Render\Element;
use Drupal\sdc\Exception\InvalidComponentException;
use Drupal\sdc\Plugin\Component;
use JsonSchema\Validator;

class ComponentValidator {

  /**
   * Validates the component metadata file.
   *
   * A valid component metadata file can be validated against the
   * metadata-author.schema.json, plus the ability of classes and interfaces
   * in the `type` property.
   *
   * @return bool
   *   TRUE if the component is valid.
   *
   * @throws \Drupal\sdc\Exception\InvalidComponentException
   */
  public function validateDefinition(array $definition): bool {
    // Detect the props with a type class, and validate that the class exists.
    $props = $definition['schemas']['props']['properties'] ?? [];
    $classes_per_prop = $this->getClassProps($definition['schemas']['props'] ?? []);
    $missing_class_errors = [];
    foreach ($props as $prop_name => $prop_def) {
      // For each possible type, check if it is a class.
      $class_types = $classes_per_prop[$prop_name] ?? [];
      $missing_classes = array_filter($class_types, static fn(string $class) => !class_exists($class) && !interface_exists($class));
      $missing_class_errors = [
        ...$missing_class_errors,
        ...array_map(
          static fn(string $class) => sprintf('Unable to find class/interface "%s" specified in the prop "%s" for the component "%s".', $class, $prop_name, $definition['id']),
          $missing_classes
        ),
      ];

    }
    // Remove the non JSON Schema types for validation down below.
    $definition['schemas']['props'] = $this->nullifyClassPropsSchema(
      $definition['schemas']['props'],
      $classes_per_prop
    );

    $validator = new Validator();
    $definition_object = Validator::arrayToObjectRecursive($definition);
    $validator->validate(
      $definition_object,
      (object) ['$ref' => 'file://' . dirname(__DIR__) . '/metadata-full.schema.json']
    );
    if (empty($missing_class_errors) && $validator->isValid()) {
      return TRUE;
    }
    $message_parts = array_map(
      static fn(array $error): string => sprintf("[%s] %s", $error['property'], $error['message']),
      $validator->getErrors()
    );
    $message_parts = [
      ...$message_parts,
      ...$missing_class_errors,
    ];
    $message = implode("/n", $message_parts);
    // Log the message.
    throw new InvalidComponentException($message);
  }

  /**
   * Validates that the props provided to the component.
   *
   * Valid props are compliant with the schema definition in the component
   * metadata file.
   *
   * @param array $context
   *   The Twig context that contains the prop data.
   * @param \Drupal\sdc\Plugin\Component $component
   *   The component to validate the props against.
   *
   * @return bool
   *   TRUE if the props adhere to the component definition.
   *
   * @throws \Drupal\sdc\Exception\InvalidComponentException
   */
  public function validateProps(array $context, Component $component): bool {
    $component_id = $component->getPluginId();
    $props_schema = $component->getMetadata()->getSchemas()['props'] ?? [];
    $prop_names = array_keys($props_schema['properties'] ?? []);
    $props_raw = array_intersect_key($context, array_flip($prop_names));
    // Validator::arrayToObjectRecursive stringifies the props using the JSON
    // encoder. Before that happens, we want to validate classes. Once the
    // classes are validated, we remove them as potential problems for the JSON
    // Schema validation.
    [
      $props_schema,
      $props_raw,
    ] = $this->validateClassProps($props_schema, $props_raw, $component_id);
    $props_schema = Validator::arrayToObjectRecursive($props_schema);
    $props = Validator::arrayToObjectRecursive($props_raw);
    $throw_error = static function (array $errors) {
      $message_parts = array_map(
        static fn(array $error): string => sprintf("[%s] %s", $error['property'], $error['message']),
        $errors
      );
      $message = implode("/n", $message_parts);
      throw new InvalidComponentException($message);
    };
    $validator = new Validator();
    $validator->validate($props, $props_schema);
    $validator->getErrors();
    if ($validator->isValid()) {
      return TRUE;
    }
    // Dismiss type errors if the prop received a render array.
    $errors = array_filter(
      $validator->getErrors(),
      function (array $error) use ($context): bool {
        if (($error['constraint'] ?? '') !== 'type') {
          return TRUE;
        }
        return !$this->isRenderArray($context[$error['property']] ?? NULL);
      }
    );
    if (empty($errors)) {
      return TRUE;
    }
    $throw_error($errors);
  }

  /**
   * Checks if a candidate is a render array.
   *
   * @param mixed $candidate
   *   The candidate.
   *
   * @return bool
   *   TRUE if it's a render array. FALSE otherwise.
   */
  protected function isRenderArray($candidate): bool {
    if (!is_array($candidate)) {
      return FALSE;
    }
    $children_keys = Element::children($candidate);
    if (empty($children_keys)) {
      // This is a render array, if there is at least one non-children element.
      return !empty($candidate);
    }
    foreach ($children_keys as $key) {
      if (!$this->isRenderArray($candidate[$key] ?? NULL)) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Validates the props that are not JSON Schema.
   *
   * This validates that the props are instances of the class/interface declared
   * in the metadata file.
   *
   * @param array $props_schema
   *   The schema for all the props in the component.
   * @param array $props_raw
   *   The props provided to the component.
   * @param string $component_id
   *   The component ID. Used for error reporting.
   *
   * @return array
   *   A tuple containing the new $props_schema and the new $props_raw. We
   *   mutate the schema to be `type: null` and the prop value to be `NULL` for
   *   the props that are validated as class objects. This is done so these
   *   props can pass validation later on when validating against the JSON
   *   Schema. We can do this because we have already validated these props
   *   manually.
   *
   * @throws \Drupal\sdc\Exception\InvalidComponentException
   */
  private function validateClassProps(array $props_schema, array $props_raw, string $component_id): array {
    $error_messages = [];
    $classes_per_prop = $this->getClassProps($props_schema);
    foreach ($props_schema['properties'] as $prop_name => $prop_def) {
      $class_types = $classes_per_prop[$prop_name] ?? [];
      $prop = $props_raw[$prop_name] ?? NULL;
      if (empty($class_types) || is_null($prop)) {
        continue;
      }
      $is_valid = array_reduce(
        $class_types,
        static fn(bool $valid, string $class_name) => $valid || $prop instanceof $class_name,
        FALSE
      );
      if (!$is_valid) {
        $error_messages[] = sprintf(
          'Data provided to prop "%s" for component "%s" is not a valid instance of "%s"',
          $prop_name,
          $component_id,
          implode(', ', $class_types),
        );
      }
      // Remove the non JSON Schema types for later JSON Schema validation.
      $props_raw[$prop_name] = NULL;
    }
    $props_schema = $this->nullifyClassPropsSchema($props_schema, $classes_per_prop);
    if (!empty($error_messages)) {
      $message = implode("/n", $error_messages);
      throw new InvalidComponentException($message);
    }
    return [$props_schema, $props_raw];
  }

  private function getClassProps(array $props_schema): array {
    $classes_per_prop = [];
    foreach ($props_schema['properties'] as $prop_name => $prop_def) {
      $type = $prop_def['type'] ?? 'null';
      $types = is_string($type) ? [$type] : $type;
      // For each possible type, check if it is a class.
      $class_types = array_filter($types, static fn(string $type) => !in_array(
        $type,
        ['array', 'boolean', 'integer', 'null', 'number', 'object', 'string']
      ));
      $classes_per_prop[$prop_name] = $class_types;
    }
    return array_filter($classes_per_prop);
  }

  /**
   * Utility method to ensure the schema for class props is set to 'null'.
   *
   * @param array $schema_props
   *   The schema for all the props.
   * @param array $classes_per_prop
   *   Associative array that associates prop names with their prop classes.
   *
   * @return array
   *   The new schema.
   */
  private function nullifyClassPropsSchema(array $schema_props, array $classes_per_prop): array {
    foreach ($schema_props['properties'] as $prop_name => $prop_def) {
      $class_types = $classes_per_prop[$prop_name] ?? [];
      // Remove the non JSON Schema types for later JSON Schema validation.
      $types = $prop_def['type'] ?? ['null'];
      $types = is_string($types) ? [$types] : $types;
      $types = array_diff($types, $class_types);
      $types = empty($types) ? ['null'] : $types;
      $schema_props['properties'][$prop_name]['type'] = $types;
    }
    return $schema_props;
  }

}
